"""
Meta Tuner Class: Used to optimize across a set of models:
- selecting intelligently the order of functions to optimize
Current implementation: Bare Metal functionality for testing.
ToDo: Improve code with better config management and remove hardcoded parameters
"""
from dataclasses import dataclass
from mango.domain.domain_space import domain_space
from mango.optimizer.bayesian_learning import BayesianLearning
from scipy.stats._distn_infrastructure import rv_frozen
import numpy as np
from tqdm.auto import tqdm
import random


## setting warnings to ignore for now
import warnings
warnings.filterwarnings('ignore')

class MetaTuner:
    @dataclass
    class Config:
        n_iter: int = 20
        n_init: int = 2

    def __init__(self,
                 param_dict_list,
                 objective_list,
                 **kwargs):
        self.param_dict_list = param_dict_list
        self.objective_list = objective_list
        self.config = MetaTuner.Config(**kwargs)
        # adjustable parameters
        self.num_of_iterations = self.config.n_iter
        self.initial_random = self.config.n_init

        #list of GPR for each objective
        self.gpr_list = []

        # store the results of MetaTuner
        self.results = dict()

        #batch size of the entire metaTuner
        self.batch_size = 1

        #batch size per function to select
        self.obj_batch_size = 1



        #use to see the info and fix the seeds metaTuner is running
        self.debug = False

        self.surrogate = None # used to test different kernel functions

        #stores the index of evaluated objectives
        self.objectives_evaluated = []

        #stores the values of objectives in the order they are evaluated
        self.objective_values_list = []

        self.seed = 0

        #rate of decay of exploration probability
        self.decay_rate = 0.90

        #The initial exploration probability
        self.exploration_rate = 1.0
        self.exploration_rate_min = 0.1

        #record which classifier/gp was used last
        self.last_used = 0

        self.store_domain = {}
        self.first_itr = {}

    def run(self):
        #run the metaTuner
        self.results = self.runExponentialTuner()
        return self.results

    @staticmethod
    def calculateDomainSize(param_dict):
        """
           Calculating the domain size to be explored for finding
           optimum of bayesian optimizer
        """
        # Minimum and maximum domain size
        domain_min = 5000
        domain_max = 30000

        domain_size = 1

        for par in param_dict:
            if isinstance(param_dict[par], rv_frozen):
                distrib = param_dict[par]
                loc, scale = distrib.args
                min_scale = 1
                scale = int(scale)
                if scale < min_scale:
                    scale = min_scale

                domain_size = domain_size * scale * 50

            elif isinstance(param_dict[par], range):
                domain_size = domain_size * len(param_dict[par])

            elif isinstance(param_dict[par], list):
                domain_size = domain_size * len(param_dict[par])

        if domain_size < domain_min:
            domain_size = domain_min

        if domain_size > domain_max:
            domain_size = domain_max

        return domain_size

    #return the max value of Y from current evaluations
    def get_max_y_value(self, Y_dict_array):
        values = Y_dict_array[0]

        for k in range(len(self.objective_list) - 1):
            values = np.vstack((values,Y_dict_array[k+1]))

        max_val = np.max(values)

        return max_val

    def runExponentialTuner(self):
        """
        Steps:
        1-Create DS obj for each obj of param_dict_list
        2-Sample randomly from the DS objects and evaluate the objective functions.
        3- Now use the GPR for each objective to select next batch
        4- Select the best values based on fxn(surrogate) from GPR.
        5- Modify the surrogate selections so as to avoid getting struck.
        """

        results = dict()

        #random params tried
        results['random_params'] = []

        #random objective values
        results['random_params_objective'] = []

        #random objective function id
        results['random_objective_fid'] = []

        results['params_tried'] = []
        results['objective_values'] = []
        results['objective_fid'] = []


        #print('*** Entering Metatuner *** ')
        num_of_random = self.initial_random

        ds = []
        for i in self.param_dict_list:
            domain_size = self.calculateDomainSize(i)
            ds.append(domain_space(i,domain_size))

        #dict of list for each obj function
        X_dict_list = {}
        Y_dict_list = {}

        #dict of array for each obj function
        X_dict_array = {}
        Y_dict_array = {}

        #stores the maximum value of the objective function for each objective
        Y_dict_array_max = {}

        #randomly evaluate the initial points for objectives
        for i in range(len(self.param_dict_list)):

            #used later
            self.first_itr[i] = True

            if self.debug:
                #fixing random seeds to reproduce the results for debugging
                np.random.seed(self.seed)
                random.seed(self.seed)

            ds_i = ds[i]

            #make sure not all values in y are same
            done = False

            #count no of times random repeated
            num_times = 0
            while not done:
                random_hyper_parameters = ds_i.get_random_sample(num_of_random)
                y_list = self.objective_list[i](random_hyper_parameters)
                #print(i, random_hyper_parameters, y_list)
                num_times = num_times + 1

                #10 times done still there is a tie
                if num_times ==10:
                    done = True
                    y_list[0] = y_list[0] - 0.001

                for ind in range(num_of_random):
                    if y_list[0]!=y_list[ind]:
                        done = True

            #print('Random_hyper_parameters:',random_hyper_parameters, y_list)
            x_list = random_hyper_parameters

            #print(i, random_hyper_parameters, x_list, y_list)

            X_dict_list[i] =[]
            X_dict_list[i].append(x_list)

            Y_dict_list[i] =[]
            Y_dict_list[i].append(y_list)

            self.objective_values_list += y_list

            results['random_params'] = results['random_params'] + x_list
            results['random_params_objective'] = results['random_params_objective'] + y_list
            results['random_objective_fid'] = results['random_objective_fid'] + [i]*len(x_list)


            #x_array2 = ds_i.convert_GP_space(random_hyper_parameters)
            #x_array = ds_i.convert_to_gp(random_hyper_parameters)

            x_array = ds_i.convert_GP_space(random_hyper_parameters)

            X_dict_array[i] = x_array

            #print(x_array2)
            #print(x_array)

            y_array = np.array(y_list).reshape(len(y_list),1)
            Y_dict_array[i] = y_array

            #the random ones are added as it is
            Y_dict_array_max[i] = y_array

        #Initialize the number of Optimizers
        Optimizer_list = []

        for i in range(len(self.objective_list)):
            Optimizer_i = BayesianLearning(surrogate=self.surrogate)
            Optimizer_i.domain_size = ds[i].domain_size
            Optimizer_list.append(Optimizer_i)

            #print('Domain Size:',ds[i].domain_size)

        # Storing optimizers attributes: iteration count and modified exploration of external surrogate
        Optimizer_exploration = []
        Optimizer_iteration = []
        for i in range(len(self.objective_list)):
            Optimizer_exploration.append(1.0)
            Optimizer_iteration.append(1.0)

        results['params_tried'] = results['random_params']
        results['objective_values'] = results['random_params_objective']
        results['objective_fid'] = results['random_objective_fid']

        #print(Optimizer_exploration)

        #Now run the optimization iterations
        pbar = tqdm(range(self.num_of_iterations))

        for itr in pbar:
            #next values of x returned from individual function
            #Values in x are dependent on types of param dict, so using a list
            x_values_list = []

            #Next promising surrogate values for each objective: used to select functions
            s_values_array = np.empty((0,1), float)

            #keeping track of objective indices: Maps the x values to its objective function
            x_obj_indices = []

            #Next promising surrogate values for each objective in list form
            s_values_list = []

            #sample individual domains and evaluate surrogate functions.
            #we get the next promising samples along with the surrogate function values
            max_val_y = self.get_max_y_value(Y_dict_array)

            #In GPs this value is used to scale the Y values
            max_val_y_scaled = max_val_y
            #In this setting the default GP performs reasonably well
            if max_val_y_scaled<1.0:
                max_val_y_scaled = 1.0

            for j in range(len(ds)):
                if self.debug:
                    #fixing random seeds to reproduce the results for debugging
                    np.random.seed(self.seed)
                    random.seed(self.seed)

                #X_domain_np = ds[j].sample_gp_space()

                if self.first_itr[j]:
                #if j==self.last_used or self.first_itr[j]:
                    # Domain Space
                    domain_list = ds[j].get_domain()
                    self.store_domain[j] = ds[j].convert_GP_space(domain_list)
                    #self.store_domain[j] = ds[j].sample_gp_space()
                    self.first_itr[j] = True
                    #print('doing domain sampling')


                X_domain_np = self.store_domain[j]

                #next batch of x for this objective along with its surrogate value
                X_next_batch, surr_value, surr_value_ext, u_values = Optimizer_list[j].get_next_batch_MetaTuner(X_dict_array[j],Y_dict_array[j]/max_val_y_scaled,X_domain_np, self.obj_batch_size, Optimizer_exploration[j], Optimizer_iteration[j], classifier_index=j, last_used_index = j)#self.last_used

                #used for displaying and debugging
                #s_values_list.append([int(surr_value_ext[0]*100.0),int(surr_value[0]*100.0)])
                s_values_list.append([int(surr_value_ext[0]*100.0)])

                #this is used to sort the objective values
                s_value = np.array(surr_value_ext) #modified sandeep

                #s_value = np.array(surr_value_ext)
                s_value = s_value.reshape(-1, s_value.shape[0])


                s_values_array = np.append(s_values_array, s_value)
                x_values_list = x_values_list + X_next_batch

                #keep track of objective function for corresponding surrogate and x values
                for k in range(self.obj_batch_size):
                    x_obj_indices.append(j)

            #sort the surrogate values in descending order, to select the best value from them
            v_sorting_index = np.argsort(-s_values_array, axis=0)

            #now select the self.batch_size values from x_values_list based on v_sorting_index
            v_sorting_index = v_sorting_index[:self.batch_size]


            #select randomly with exploration rate else select the max surrogate

            prob_selection = random.random()

            random_selected = False

            #do the random evaluation of the functions
            if prob_selection < self.exploration_rate:
                selected_obj = random.randint(0,(len(ds)-1)) #0 and len(ds included)

                if self.exploration_rate>self.exploration_rate_min:
                    self.exploration_rate = self.exploration_rate*self.decay_rate

                else:
                    self.exploration_rate = self.exploration_rate_min

                random_selected = True

            else:
                selected_obj = v_sorting_index[0]

            #keep track of objective indices selected in current iteration
            loc_indices = []
            curr_x_next_np = x_values_list[selected_obj]

            #convert this into the parameter space for scheduling
            #see the function index for this x value
            index = selected_obj

            # updating last used classifier
            self.last_used  = index

            #keep track of local indices
            loc_indices.append(index)

            #keep track of indices in a global datastr for visualization of function selection
            self.objectives_evaluated.append(index)

            curr_x_next = ds[index].convert_PS_space(curr_x_next_np)

            #run the next curr_x_next value for the objective function
            y_list = self.objective_list[index](curr_x_next)


            self.objective_values_list += y_list

            results['params_tried'] = results['params_tried'] + curr_x_next
            results['objective_values'] = results['objective_values'] + y_list
            results['objective_fid'] = results['objective_fid'] + [index]

            curr_y_array = np.array(y_list).reshape(len(y_list), 1)
            #append the curr_x_next_np, curr_x_next, y_list to appropriate datastructures for book keeping

            X_dict_array[index] = np.vstack((X_dict_array[index], curr_x_next_np))
            Y_dict_array[index] = np.vstack((Y_dict_array[index], curr_y_array))

            Y_dict_array_max[index] = np.vstack((Y_dict_array_max[index], np.max(Y_dict_array[index])))

            #scale the exploration of objectives that are not selected for others make their exploration to 1
            for i in range(len(self.objective_list)):
                if i in loc_indices:
                    Optimizer_exploration[i] = 1.0
                    Optimizer_iteration[i] += 1.0
                else:
                    Optimizer_exploration[i] = Optimizer_exploration[i] + 1.1*self.exploration_rate


            pbar.set_description(": Best score: %s" % max_val_y)

            #print(itr, s_values_list, Optimizer_iteration, Optimizer_exploration, max_val_y)#, Y_dict_array[selected_obj])

            #print(itr, s_values_list, random_selected, self.exploration_rate, prob_selection, Optimizer_iteration , max_val_y)

            #print(itr, s_values_list, random_selected, Optimizer_iteration , int(self.exploration_rate*100.0), max_val_y)

            self.X_dict_array = X_dict_array
            self.Y_dict_array = Y_dict_array
            self.Y_dict_array_max = Y_dict_array_max
            self.ds = ds

        self.X_dict_array = X_dict_array
        self.Y_dict_array = Y_dict_array
        self.Y_dict_array_max = Y_dict_array_max
        self.ds = ds

        index_best = np.argmax(results['objective_values'])
        results['best_objective'] = results['objective_values'][index_best]
        results['best_params'] = results['params_tried'][index_best]
        results['best_objective_fid'] = results['objective_fid'][index_best]

        return results



import abc


class BasePredictor(object):

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def get_next_batch(self, X,Y,X_tries):
    """
    Gives the batch of next suggestions to try for the algorithm
    """
    raise NotImplementedError



"""
Main Tuner Class which uses other abstractions.
General usage is to find the optimal hyper-parameters of the classifier
"""

import copy
from dataclasses import dataclass, field
import logging
import random
from typing import Callable

from mango.domain.domain_space import domain_space
from mango.optimizer.bayesian_learning import BayesianLearning
from scipy.stats._distn_infrastructure import rv_frozen

from tqdm.auto import tqdm
import numpy as np

# setting warnings to ignore for now
import warnings

warnings.filterwarnings('ignore')

_logger = logging.getLogger(__name__)


class Tuner:
    @dataclass
    class Config:
        domain_size: int = None
        initial_random: int = 2
        initial_custom: dict = None
        num_iteration: int = 20
        batch_size: int = 1
        optimizer: str = 'Bayesian'
        parallel_strategy: str = 'clustering'
        surrogate: object = None  # used to test different kernel functions
        valid_optimizers = ['Bayesian', 'Random']
        valid_parallel_strategies = ['penalty', 'clustering']
        alpha: float = 2.0
        exploration: float = 1.0
        exploration_decay: float = 0.9
        exploration_min: float = 0.1
        fixed_domain: bool = False
        early_stopping: Callable = None

        def __post_init__(self):
            if self.optimizer not in self.valid_optimizers:
                raise ValueError(f'optimizer: {self.optimizer} is not valid, should be one of {self.valid_optimizers}')
            if self.parallel_strategy not in self.valid_parallel_strategies:
                raise ValueError(
                    f'parallel strategy: {self.parallel_strategy} is not valid, should be one of {self.valid_parallel_strategies}')

        @property
        def is_bayesian(self):
            return self.optimizer == 'Bayesian'

        @property
        def is_random(self):
            return self.optimizer == 'Random'

        @property
        def strategy_is_penalty(self):
            return self.parallel_strategy == 'penalty'

        @property
        def strategy_is_clustering(self):
            return self.parallel_strategy == 'clustering'

        def early_stop(self, results):
            if self.early_stopping is None:
                return False

            results = copy.deepcopy(results)
            return self.early_stopping(results)

    def __init__(self, param_dict, objective, conf_dict=None):

        self.param_dict = param_dict
        self.objective_function = objective
        self.maximize_objective = True

        if conf_dict is None:
            conf_dict = {}

        self.config = Tuner.Config(**conf_dict)

        if self.config.domain_size is None:
            self.config.domain_size = self.calculateDomainSize(self.param_dict)

        # overwrite batch size if given as a property of objective function
        if hasattr(objective, 'batch_size'):
            self.config.batch_size = objective.batch_size

        # save domain size
        self.ds = domain_space(self.param_dict, self.config.domain_size)

        # stores the results of using the tuner
        self.results = dict()

    @staticmethod
    def calculateDomainSize(param_dict):
        """
           Calculating the domain size to be explored for finding
           optimum of bayesian optimizer
        """
        # Minimum and maximum domain size
        domain_min = 50000
        domain_max = 500000

        domain_size = 1

        for par in param_dict:
            if isinstance(param_dict[par], rv_frozen):
                distrib = param_dict[par]
                loc, scale = distrib.args
                min_scale = 1
                scale = int(scale)
                if scale < min_scale:
                    scale = min_scale

                domain_size = domain_size * scale * 50

            elif isinstance(param_dict[par], range):
                domain_size = domain_size * len(param_dict[par])

            elif isinstance(param_dict[par], list):
                domain_size = domain_size * len(param_dict[par])

        if domain_size < domain_min:
            domain_size = domain_min

        if domain_size > domain_max:
            domain_size = domain_max

        return domain_size

    def run(self):
        if self.config.is_bayesian:
            self.results = self.runBayesianOptimizer()
        elif self.config.is_random:
            self.results = self.runRandomOptimizer()
        else:
            raise ValueError("Unknown Optimizer %s" % self.config.optimizer)

        return self.results

    def maximize(self):
        return self.run()

    def minimize(self):
        self.maximize_objective = False
        return self.run()


    def run_initial(self):
        if self.config.initial_custom is not None:
            X_tried = copy.deepcopy(self.config.initial_custom)
            X_list, Y_list = self.runUserObjective(X_tried)
        else:
            # getting first few random values
            X_tried = self.ds.get_random_sample(self.config.initial_random)
            X_list, Y_list = self.runUserObjective(X_tried)

            # in case initial random results are invalid try different samples
            n_tries = 1
            while len(Y_list) < self.config.initial_random and n_tries < 3:
                X_tried2 = self.ds.get_random_sample(self.config.initial_random - len(Y_list))
                X_list2, Y_list2 = self.runUserObjective(X_tried2)
                X_tried2.extend(X_tried2)
                X_list = np.append(X_list, X_list2)
                Y_list = np.append(Y_list, Y_list2)
                n_tries += 1

            if len(Y_list) == 0:
                raise ValueError("No valid configuration found to initiate the Bayesian Optimizer")
        return X_list, Y_list, X_tried

    def runBayesianOptimizer(self):
        results = dict()

        X_list, Y_list, X_tried = self.run_initial()


        # evaluated hyper parameters are used
        X_init = self.ds.convert_GP_space(X_list)
        Y_init = Y_list.reshape(len(Y_list), 1)

        # setting the initial random hyper parameters tried
        results['random_params'] = X_list
        results['random_params_objective'] = Y_list

        Optimizer = BayesianLearning(surrogate=self.config.surrogate,
                                     alpha=self.config.alpha,
                                     domain_size=self.config.domain_size)

        X_sample = X_init
        Y_sample = Y_init

        hyper_parameters_tried = X_tried
        objective_function_values = Y_list
        surrogate_values = Y_list

        x_failed_evaluations = np.array([])

        domain_list = self.ds.get_domain()
        X_domain_np = self.ds.convert_GP_space(domain_list)
        context = None

        # running the iterations
        pbar = tqdm(range(self.config.num_iteration))
        for i in pbar:

            # adding a Minimum exploration to explore independent of UCB
            if random.random() < self.config.exploration:
                random_parameters = self.ds.get_random_sample(self.config.batch_size)
                X_next_batch = self.ds.convert_GP_space(random_parameters)

                if self.config.exploration > self.config.exploration_min:
                    self.config.exploration = self.config.exploration * self.config.exploration_decay

            elif self.config.strategy_is_penalty:
                X_next_batch = Optimizer.get_next_batch(X_sample, Y_sample, X_domain_np,
                                                        batch_size=self.config.batch_size)
            elif self.config.strategy_is_clustering:
                X_next_batch = Optimizer.get_next_batch_clustering(X_sample, Y_sample, X_domain_np,
                                                                   batch_size=self.config.batch_size)
            else:
                # assume penalty approach
                X_next_batch = Optimizer.get_next_batch(X_sample, Y_sample, X_domain_np,
                                                        batch_size=self.config.batch_size)

            # Scheduler
            X_next_PS = self.ds.convert_PS_space(X_next_batch)

            # if all the xs have failed before, replace them with random sample
            # as we will not get any new information otherwise
            if all(x in x_failed_evaluations for x in X_next_PS):
                X_next_PS = self.ds.get_random_sample(self.config.batch_size)

            # Evaluate the Objective function
            X_next_list, Y_next_list = self.runUserObjective(X_next_PS)

            # keep track of all parameters that failed
            x_failed = [x for x in X_next_PS if x not in X_next_list]
            x_failed_evaluations = np.append(x_failed_evaluations, x_failed)

            if len(Y_next_list) == 0:
                # no values returned
                # this is problematic if domain is small and same value is tried again in the next iteration as the optimizer would be stuck
                continue

            Y_next_batch = Y_next_list.reshape(len(Y_next_list), 1)
            # update X_next_batch to successfully evaluated values
            X_next_batch = self.ds.convert_GP_space(X_next_list)

            # update the bookeeping of values tried
            hyper_parameters_tried = np.append(hyper_parameters_tried, X_next_list)
            objective_function_values = np.append(objective_function_values, Y_next_list)
            surrogate_values = np.append(surrogate_values, Optimizer.surrogate.predict(X_next_batch))

            # Appending to the current samples
            X_sample = np.vstack((X_sample, X_next_batch))
            Y_sample = np.vstack((Y_sample, Y_next_batch))

            # referesh domain if not fixed
            if not self.config.fixed_domain:
                domain_list = self.ds.get_domain()
                X_domain_np = self.ds.convert_GP_space(domain_list)

            results['params_tried'] = hyper_parameters_tried
            results['objective_values'] = objective_function_values
            results['surrogate_values'] = surrogate_values

            results['best_objective'] = np.max(results['objective_values'])
            results['best_params'] = results['params_tried'][np.argmax(results['objective_values'])]
            if self.maximize_objective is False:
                results['objective_values'] = -1 * results['objective_values']
                results['best_objective'] = -1 * results['best_objective']

            pbar.set_description("Best score: %s" % results['best_objective'])

            # check if early stop criteria has been met
            if self.config.early_stop(results):
                _logger.info('Early stopping criteria satisfied')
                break

        # saving the optimizer and ds in the tuner object which can save the surrogate function and ds details
        self.Optimizer = Optimizer

        return results

    def runRandomOptimizer(self):
        results = dict()

        X_sample_list = []
        Y_sample_list = []

        batch_size = self.config.batch_size
        n_iterations = self.config.num_iteration
        random_hyper_parameters = self.ds.get_random_sample(n_iterations * batch_size)

        # running the iterations
        pbar = tqdm(range(0, len(random_hyper_parameters), batch_size))
        for idx in pbar:
            # getting batch by batch random values to try
            batch_hyper_parameters = random_hyper_parameters[idx:idx + batch_size]
            X_list, Y_list = self.runUserObjective(batch_hyper_parameters)

            X_sample_list = np.append(X_sample_list, X_list)
            Y_sample_list = np.append(Y_sample_list, Y_list)

            results['params_tried'] = X_sample_list
            results['objective_values'] = Y_sample_list

            results['best_objective'] = np.max(results['objective_values'])
            results['best_params'] = results['params_tried'][np.argmax(results['objective_values'])]
            if self.maximize_objective is False:
                results['objective_values'] = -1 * results['objective_values']
                results['best_objective'] = -1 * results['best_objective']

            pbar.set_description("Best score: %s" % results['best_objective'])

            # check if early stop criteria has been met
            if self.config.early_stop(results):
                _logger.info('Early stopping criteria satisfied')
                break

        return results

    def runUserObjective(self, X_next_PS):
        # initially assuming entire X_next_PS is evaluated and returned results are only Y values
        X_list_evaluated = X_next_PS
        results = self.objective_function(X_next_PS)
        Y_list_evaluated = results

        # if result is a tuple, then there is possibility that partial values are evaluated
        if isinstance(results, tuple):
            X_list_evaluated, Y_list_evaluated = results

        X_list_evaluated = np.array(X_list_evaluated)
        if self.maximize_objective is False:
            Y_list_evaluated = -1 * np.array(Y_list_evaluated)
        else:
            Y_list_evaluated = np.array(Y_list_evaluated)

        return X_list_evaluated, Y_list_evaluated



from collections.abc import Mapping, Iterable
import warnings

import numpy as np
from sklearn.utils import check_random_state
from sklearn.utils.random import sample_without_replacement
from sklearn.model_selection import ParameterGrid


class BatchParameterSampler:

    def __init__(self, param_distribution, n_iter, *, random_state=None):
        if not isinstance(param_distribution, (Mapping)):
            raise TypeError('Parameter distribution is not a dict or '
                            '({!r})'.format(param_distribution))

        dist = param_distribution
        for key in dist:
            if (not isinstance(dist[key], Iterable)
                    and not hasattr(dist[key], 'rvs')):
                raise TypeError('Parameter value is not iterable '
                                'or distribution (key={!r}, value={!r})'
                                .format(key, dist[key]))
        self.n_iter = n_iter
        self.random_state = random_state
        self.dist = param_distribution

    def _is_all_lists(self):
        return all(not hasattr(v, "rvs") for v in self.dist.values())

    def __iter__(self):
        rng = check_random_state(self.random_state)

        # if all distributions are given as lists, we want to sample without
        # replacement
        if self._is_all_lists():
            # look up sampled parameter settings in parameter grid
            param_grid = ParameterGrid(self.dist)
            grid_size = len(param_grid)
            n_iter = self.n_iter

            if grid_size < n_iter:
                warnings.warn(
                    'The total space of parameters %d is smaller '
                    'than n_iter=%d. Running %d iterations. For exhaustive '
                    'searches, use GridSearchCV.'
                    % (grid_size, self.n_iter, grid_size), UserWarning)
                n_iter = grid_size
            for i in sample_without_replacement(grid_size, n_iter,
                                                random_state=rng):
                yield param_grid[i]

        else:
            # Always sort the keys of a dictionary, for reproducibility
            items = sorted(self.dist.items())
            samples = []
            keys = []
            for k, v in items:
                keys.append(k)
                if hasattr(v, "rvs"):
                    samples.append(v.rvs(random_state=rng, size=self.n_iter))
                else:
                    # forcing object dtype to avoid conversion of mixed datatypes
                    v = np.array(v, dtype=object)
                    samples.append(rng.choice(v, size=self.n_iter))

            for sample in zip(*samples):
                yield dict(zip(keys, sample))

    def __len__(self):
        """Number of points that will be sampled."""
        if self._is_all_lists():
            grid_size = len(ParameterGrid(self.dist))
            return min(self.n_iter, grid_size)
        else:
            return self.n_iter




import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern
import math
from sklearn.cluster import KMeans

from .base_predictor import BasePredictor

"""
Bayesian Learning optimizer
- We will use the Gaussian Prior by default with the Matern Kernel
"""


class BayesianLearning(BasePredictor):

    def __init__(self, surrogate=None, alpha=None, domain_size=1000):
        # initialzing some of the default values
        # The default surrogate function is gaussian_process with matern kernel
        if surrogate is None:
            self.surrogate = GaussianProcessRegressor(kernel=Matern(nu=2.5),
                                                      n_restarts_optimizer=10,
                                                      # FIXME:check if we should be passing this
                                                      # random state
                                                      random_state=1,
                                                      normalize_y=True)
        else:
            self.surrogate = surrogate

        # keep track of the iteration counts
        self.iteration_count = 0

        # The size of the exploration domain, default to 1000
        self.domain_size = domain_size

        self.alpha = alpha

    def Upper_Confidence_Bound_Remove_Duplicates(self, X, X_Sample, batch_idx):
        """
            Check if the returned index value is already present in X_Sample
        """
        mu, sigma = self.surrogate.predict(X, return_std=True)
        mu = mu.reshape(mu.shape[0], 1)
        sigma = sigma.reshape(sigma.shape[0], 1)

        # use fixed alpha if given
        if self.alpha is not None:
            alpha = self.alpha
        else:
            alpha_inter = self.domain_size * (self.iteration_count) * (self.iteration_count) * math.pi * math.pi / (
                    6 * 0.1)

            if alpha_inter == 0:
                raise ValueError('alpha_inter is zero in Upper_Confidence_Bound')

            alpha = 2 * math.log(alpha_inter)  # We have set delta = 0.1
            alpha = math.sqrt(alpha)

        if batch_idx == 0:
            exploration_factor = alpha
        else:
            tolerance = 1e-6
            sigma_inv_sq = 1.0 / (tolerance + (sigma * sigma))  # tolerance is used to avoid the divide by zero error
            C = 8 / (np.log(1 + sigma_inv_sq))
            beta = np.exp(2 * C) * alpha
            beta = np.sqrt(beta)
            exploration_factor = beta

        Value = mu + exploration_factor * sigma

        return self.remove_duplicates(X, X_Sample, mu, Value)

    def Get_Upper_Confidence_Bound(self, X):
        """
            Returns the acqutition function
        """
        mu, sigma = self.surrogate.predict(X, return_std=True)
        mu = mu.reshape(mu.shape[0], 1)
        sigma = sigma.reshape(sigma.shape[0], 1)

        if self.alpha is not None:
            exploration_factor = self.alpha
        else:
            alpha_inter = self.domain_size * (self.iteration_count) * (self.iteration_count) * math.pi * math.pi / (
                    6 * 0.1)
            if alpha_inter == 0:
                raise ValueError('alpha_inter is zero in Upper_Confidence_Bound')
            alpha = 2 * math.log(alpha_inter)  # We have set delta = 0.1
            alpha = math.sqrt(alpha)

            exploration_factor = alpha

        Value = mu + exploration_factor * sigma

        return Value

    """
    Returns the most optmal x along with mean value from the domain of x and making sure it is not a Duplicate (depending on closeness)
    used in batch setting: As mean is also returned
    """

    def remove_duplicates(self, X, X_Sample, mu, Value):
        # print('*'*200)
        v_sorting_index = np.argsort(-Value, axis=0)
        index = 0
        # go through all the values in X_Sample and check if anyvalue is close
        # to the optimal x value, if yes, don't consider this optimal x value

        while index < v_sorting_index.shape[0]:
            x_optimal = X[v_sorting_index[index]]

            # check if x_optimal is in X_Sample
            check_closeness = self.closeness(x_optimal, X_Sample)

            if check_closeness == False:  # No close element to x_optimal in X_Sample
                break

                # we will look for next optimal value to try
            else:
                index = index + 1

        # If entire domain is same to the already selected samples, we will just pick the best by value then
        if (index == v_sorting_index.shape[0]):
            index = 0

        return X[v_sorting_index[index]], mu[v_sorting_index[index]]

    """
    Returns the most optmal x only from the domain of x and making sure it is not a Duplicate (depending on closeness)
    Intended for usage in serial and clustering setting: As no mean is also returned, and no hullicination is considered
    """

    def remove_duplicates_serial(self, X, X_Sample, Value):
        # print('*'*200)
        v_sorting_index = np.argsort(-Value, axis=0)
        index = 0
        # go through all the values in X_Sample and check if anyvalue is close
        # to the optimal x value, if yes, don't consider this optimal x value

        while index < v_sorting_index.shape[0]:
            x_optimal = X[v_sorting_index[index]]

            # check if x_optimal is in X_Sample
            check_closeness = self.closeness(x_optimal, X_Sample)

            if check_closeness == False:  # No close element to x_optimal in X_Sample
                break

                # we will look for next optimal value to try
            else:
                index = index + 1

        # If entire domain is same to the already selected samples, we will just pick the best by value then
        if (index == v_sorting_index.shape[0]):
            index = 0

        return X[v_sorting_index[index]]

    def closeness(self, x_optimal, X_Sample):
        # check if x_optimal is close to X_Sample
        tolerance = 1e-3

        for i in range(X_Sample.shape[0]):
            diff = np.sum(np.absolute(X_Sample[i] - x_optimal))
            if (diff < tolerance):
                # print('Removed Duplicate')
                return True

        return False

    """
    This is the main function which returns the next batch to try along with the mean values for this batch
    """

    def get_next_batch(self, X, Y, X_tries, batch_size):
        # print('In get_next_batch')

        X_temp = X
        Y_temp = Y

        batch = []

        for idx in range(batch_size):
            self.iteration_count = self.iteration_count + 1
            self.surrogate.fit(X_temp, Y_temp)

            X_next, u_value = self.Upper_Confidence_Bound_Remove_Duplicates(X_tries, X_temp, idx)
            u_value = u_value.reshape(-1, 1)
            Y_temp = np.vstack((Y_temp, u_value))
            X_temp = np.vstack((X_temp, X_next))

            batch.append([X_next])

        batch = np.array(batch)

        batch = batch.reshape(-1, X.shape[1])
        return batch

    """
    Using clustering to select next batch
    """

    def get_next_batch_clustering(self, X, Y, X_tries, batch_size):
        # print('In get_next_batch')

        X_temp = X
        Y_temp = Y

        self.surrogate.fit(X_temp, Y_temp)
        self.iteration_count = self.iteration_count + 1

        Acquition = self.Get_Upper_Confidence_Bound(X_tries)

        if batch_size > 1:
            gen = sorted(zip(Acquition, X_tries), key=lambda x: -x[0])
            x_best_acq_value, x_best_acq_domain = (np.array(t)[:len(Acquition) // 4]
                                                   for t in zip(*gen))

            # Do the domain space based clustering on the best points
            kmeans = KMeans(n_clusters=batch_size, random_state=0).fit(x_best_acq_domain)
            cluster_pred_domain = kmeans.labels_.reshape(kmeans.labels_.shape[0])

            # partition the space into the cluster in X and select the best X from each space
            partitioned_space = dict()
            partitioned_acq = dict()
            for i in range(batch_size):
                partitioned_space[i] = []
                partitioned_acq[i] = []

            for i in range(x_best_acq_domain.shape[0]):
                partitioned_space[cluster_pred_domain[i]].append(x_best_acq_domain[i])
                partitioned_acq[cluster_pred_domain[i]].append(x_best_acq_value[i])

            batch = []

            for i in partitioned_space:
                x_local = partitioned_space[i]
                acq_local = partitioned_acq[i]
                acq_local = np.array(acq_local)
                x_index = np.argmax(acq_local)
                x_final_selected = x_local[x_index]
                batch.append([x_final_selected])

        else:  # batch_size ==1
            batch = []
            x_index = np.argmax(Acquition)
            x_final_selected = self.remove_duplicates_serial(X_tries, X_temp, Acquition)
            # x_final_selected = X_tries[x_index]
            batch.append([x_final_selected])

        batch = np.array(batch)
        batch = batch.reshape(-1, X.shape[1])
        return batch

    #used by MetaTuner which also returns the surrogate function values
    def remove_duplicates_MetaTuner(self, X, X_Sample, mu, Value, Value_ext):
        # print('*'*200)
        v_sorting_index = np.argsort(-Value, axis=0)
        index = 0
        # go through all the values in X_Sample and check if anyvalue is close
        # to the optimal x value, if yes, don't consider this optimal x value

        while index < v_sorting_index.shape[0]:
            x_optimal = X[v_sorting_index[index]]

            # check if x_optimal is in X_Sample
            check_closeness = self.closeness(x_optimal, X_Sample)

            if check_closeness == False:  # No close element to x_optimal in X_Sample
                break

                # we will look for next optimal value to try
            else:
                index = index + 1

        # If entire domain is same to the already selected samples, we will just pick the best by value then
        if (index == v_sorting_index.shape[0]):
            index = 0

        return X[v_sorting_index[index]], Value[v_sorting_index[index]], Value_ext[v_sorting_index[index]], mu[v_sorting_index[index]]


    """
    Used by MetaTuner by Exponentially scaling the exploration factor
    """
    def Upper_Confidence_Bound_Remove_Duplicates_MetaTuner(self, X, X_Sample, batch_size, exploration_factor_tuner, Optimizer_iteration):
        mu, sigma = self.surrogate.predict(X, return_std=True)
        mu = mu.reshape(mu.shape[0], 1)

        sigma = sigma.reshape(sigma.shape[0], 1)


        #Optimizer_iteration: Number of times this function has been selected
        alpha_inter = self.domain_size * (Optimizer_iteration) * (Optimizer_iteration) * math.pi * math.pi / (6 * 0.1)

        if alpha_inter == 0:
            print('Error: alpha_inter is zero in Upper_Confidence_Bound')

        alpha = 2 * math.log(alpha_inter)  # We have set delta = 0.1
        alpha = math.sqrt(alpha)

        if batch_size == 1:
            exploration_factor = 2.0
        else: #for both cases we don't modify exploration factor: ToDO see if this if optimal
            exploration_factor = 2.0

        #used for internal exploration
        Value = mu + (exploration_factor * sigma)

        #used for exploration among functions
        Value_ext = mu + (exploration_factor_tuner * sigma)


        return self.remove_duplicates_MetaTuner(X, X_Sample, mu, Value, Value_ext)


    """
    Function used to select the next batch by MetaTuner
    """
    def get_next_batch_MetaTuner(self, X, Y, X_tries, batch_size=3, exploration_factor =1.0, Optimizer_iteration=1.0, classifier_index=0, last_used_index=0):
        # print('In get_next_batch')

        X_temp = X
        Y_temp = Y

        batch = []

        #value of the surrogate function
        s_values = []


        #value of the external surrogate functio
        s_values_ext = []


        #value used to do external exploration among functions
        u_values = []


        for i in range(batch_size):

            # fit only if some modified
#             if classifier_index==last_used_index:
#                 self.surrogate.fit(X_temp, Y_temp)
#                 print('Doing fit of surrogate:',classifier_index)

            try:
                self.surrogate.fit(X_temp, Y_temp)

            except:
                print('*'*100)
                print(X_temp)
                print('*'*10)
                print(Y_temp)
            #print('Doing fit of surrogate:',classifier_index)

            X_next, s_value, s_value_ext, u_value = self.Upper_Confidence_Bound_Remove_Duplicates_MetaTuner(X_tries, X_temp, batch_size, exploration_factor, Optimizer_iteration)

            Optimizer_iteration = Optimizer_iteration + 1

            s_values.append(s_value[0][0])
            s_values_ext.append(s_value_ext[0][0])
            u_values.append(u_value[0][0])

            u_value = u_value.reshape(-1, 1)
            Y_temp = np.vstack((Y_temp, u_value))
            X_temp = np.vstack((X_temp, X_next))

            batch.append(X_next)

        #batch = np.array(batch)
        #s_value = np.array(s_value)
        #batch = batch.reshape(-1, X.shape[1])
        #s_value = s_value.reshape(-1, s_value.shape[0])

        return batch, s_values, s_values_ext, u_values


    """
    Get the predictions from the surrogate function
    along with the variance
    """

    def predict(self, X):
        pred_y, sigma = self.surrogate.predict(X, return_std=True)
        return pred_y, sigma

    """
    fit the optimizer on the X and Y values
    """

    def fit(self, X, Y):
        self.surrogate.fit(X, Y)



import setuptools

setuptools.setup(
    name='arm-mango',
    version='1.2.0',
    author='Arm Research',
    author_email='mohit.aggarwal@arm.com',
    description='parallel bayesian optimization over complex search spaces',
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/ARM-software/mango',
    packages=['mango', 'mango.optimizer', 'mango.domain'],
    include_package_data=True,
    package_data={
        '': ['*.cfg'],
    },
    install_requires=[
        'numpy>=1.17.0',
        'scipy>=1.4.1',
        'scikit_learn>=0.21.3',
        'tqdm>=4.36.1',
        'attrdict>=2.0.1',
        'dataclasses'
    ],
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
)



"""
Define the domain space abstractions for the Optimizer
"""

# this is used to check whether the domain space object is a of type rv_frozen.
# if the parameter type is a disribution or rv_frozen, the parameter sampler can handle it
from scipy.stats._distn_infrastructure import rv_frozen
import numpy as np
from sklearn.preprocessing import StandardScaler
from collections.abc import Iterable

from .batch_parameter_sampler import BatchParameterSampler


class domain_space():
    """
     initializer
     1) We expect a parameter dictionary as an input from which we will creating a mapping.
     2) The mapping is storing the categorical variables or discrete variables along with the discrete possible values
     3) The categorical variables are handeled as one hot encoding values
    """

    def __init__(self,
                 param_dict,
                 domain_size):
        self.param_dict = param_dict

        # the domain size to explore using the parameter sampler
        self.domain_size = domain_size

        # creating a mapping of categorical variables
        self.create_mappings()


    """
    returns the list of domain values using the parameter sampler
    The size of values is the domain_size
    """

    def get_domain(self):
        domain_list = list(BatchParameterSampler(self.param_dict, n_iter=self.domain_size))
        return domain_list

    """
    return a random sample using the ParameterSample
    """

    def get_random_sample(self, size):
        domain_list = list(BatchParameterSampler(self.param_dict, n_iter=size))
        return domain_list

    """
    Categorical values/discrete values are considered from the list of each value being str

    We will finally do one-hot-encoding of the list values.
    Here we will keep some book-keeping information, like number of different values
    and the mapping of each

    Integer values are considered from list of each value as int or from a range
    """

    def create_mappings(self):
        mapping_categorical = dict()
        mapping_int = dict()

        param_dict = self.param_dict
        for par in param_dict:
            if isinstance(param_dict[par], rv_frozen):
                # FIXME: what if the distribution generators ints , GP would convert it to float
                pass  # we are not doing anything at present, and will directly use its value for GP.

            elif isinstance(param_dict[par], range):
                mapping_int[par] = list(param_dict[par])

            elif isinstance(param_dict[par], Iterable):

                # for list with all int we are considering it as non-categorical
                try:
                    # this check takes care of numpy ints as well
                    all_int = all(x == int(x) and type(x) != bool
                                  for x in param_dict[par])
                except (ValueError, TypeError):
                    all_int = False

                if all_int:
                    mapping_int[par] = list(param_dict[par])

                # For lists with mixed type, floats or strings we consider them categorical or discrete
                else:
                    mapping_categorical[par] = list(param_dict[par])

        self.mapping_categorical = mapping_categorical
        self.mapping_int = mapping_int


    """
    convert the hyperparameters from the param_dict space to the GP space, by converting the
    categorical variables to one hotencoded, and return a numpy array which can be used to train the GP

    input is the domain_list generated using the Parameter Sampler.
    """

    def convert_GP_space(self, domain_list):
        mapping_categorical = self.mapping_categorical

        X = []
        for domain in domain_list:
            curr_x = []
            # for x in domain:
            for x in sorted(domain.keys()):
                # this value can be directly used, for int too, we will consider it as a float for GP
                if x not in mapping_categorical:
                    curr_x.append(domain[x])

                # this is a categorical variable which require special handling
                elif x in mapping_categorical:

                    size = len(mapping_categorical[x])  # total number of categories.

                    # we need to see the index where: domain[x] appears in mapping[x]
                    index = mapping_categorical[x].index(domain[x])

                    listofzeros = [0.0] * size

                    # We will set the value to one for one hot encoding
                    listofzeros[index] = 1.0

                    # expanding current list
                    curr_x = curr_x + listofzeros

            X.append(curr_x)

        X = np.array(X)

        return X

    """
    Convert from the X_gp space which is a numpy array that can be given input to the gaussian process
    to the parameter sampler space which is a list of the dict

    We have to reverse the one-hotencoded transformation of the categories to the category name
    """

    def convert_PS_space(self, X_gp):
        X_ps = []

        mapping_categorical = self.mapping_categorical
        mapping_int = self.mapping_int
        param_dict = self.param_dict

        for i in range(X_gp.shape[0]):

            curr_x_gp = X_gp[i]
            # we will create a list of dict, same as parameter samplers
            curr_x_ps = dict()

            index = 0

            # every sample is from the param_dict

            for par in sorted(param_dict.keys()):
                # for par in param_dict:
                # print('par is:',par)

                # this has to have integer values
                if par in mapping_int:
                    curr_x_ps[par] = int(round(curr_x_gp[index]))
                    index = index + 1

                # this par is a categorical variable and we need to handle it carefully
                elif par in mapping_categorical:
                    size = len(mapping_categorical[par])  # total number of categories.

                    one_hot_encoded = curr_x_gp[index:index + size]
                    category_type = np.argmax(one_hot_encoded)

                    category_type = mapping_categorical[par][category_type]

                    curr_x_ps[par] = category_type

                    # we have processed the entire one-hotencoded
                    index = index + size

                # this is a float value
                else:
                    curr_x_ps[par] = curr_x_gp[index]
                    index = index + 1

            X_ps.append(curr_x_ps)

        return X_ps
